package ga.cellular;

import laboratory.common.ga.GA;
import laboratory.common.ga.Individual;
import laboratory.common.ga.IndividualFactory;

import java.util.*;

public class CellularGA implements GA {

    private int width, height;

    private List<Individual> generation;
    private final double probabilityMutation;
    private final double sizeElite;
    private final IndividualFactory factory;
    private final Random r;     

    // equivalent of std::pair
    class Pair< T1 extends Comparable , T2 extends Comparable > implements Comparable< Pair< T1, T2> >
    {
        private T1 first;
        private T2 second;

        public Pair( T1 first, T2 second )
        {
            this.first = first;
            this.second = second;
        }

        public T1 getFirst()
        {
            return first;
        }

        public T2 getSecond()
        {
            return second;
        }

        public int compareTo(Pair<T1, T2> o) {
            return getFirst().compareTo( o.getFirst() );
            //return o.getFirst().compareTo( getFirst() );
        }
    }

    public void nextGeneration() {
        int size = generation.size();
        List<Individual> newGeneration = new ArrayList<Individual>( size );

//      , 
// ,   .

//        Pair< Double, Integer >[] p = new Pair[ size ];
//        for ( int i = 0; i != size; i++ )
//        {
//            p[i] = new Pair< Double, Integer >( generation.get(i).fitness(), i );
//        }
//        Arrays.sort( p );
//
//        TreeSet< Integer > eliteIndividuals = new TreeSet< Integer >();
//        int limit = (int)(size * sizeElite);
//        for ( int i = size - 1; i >= size - limit; --i )
//        {
//            eliteIndividuals.add( p[ i ].getSecond() );
//        }

        for ( int x = 0; x != height; ++x )
        {
            for ( int y = 0; y != width; ++y )
            {
                int cur = x * width + y;
//                if ( eliteIndividuals.contains( cur ) )
//                {
//                    newGeneration.add( generation.get(cur) );
//                    continue;
//                }
                List< Individual > nghb = getNeighbours( x, y );
                Individual best = nghb.get( 0 );
                for ( Individual i : nghb )
                {
                    if ( i.fitness() > best.fitness() )
                    {
                        best = i;                      
                    }
                }
                Individual[] cross = best.crossover( generation.get( cur ), r );
                //int idx = r.nextInt( 2 );
                Individual newIndividual = null;
                if ( generation.get( cur ).fitness() >= cross[ 0 ].fitness() && generation.get( cur ).fitness() >= cross[ 1 ].fitness() )
                    newIndividual = generation.get( cur );
                else
                    newIndividual = cross[ 0 ].fitness() > cross[ 1 ].fitness()? cross[ 0 ] : cross[ 1 ];                    
                
                newGeneration.add( newIndividual );
            }
        }

        for ( int i = 0; i != size; ++i )
        {
            boolean shouldMutate = r.nextDouble() <= probabilityMutation;
            generation.set( i, shouldMutate ? newGeneration.get(i).mutate( r ) : newGeneration.get(i) );
        }
        //generation = newGeneration;                
    }

    public List<Individual> getGeneration() {
        ArrayList< Individual > g = new ArrayList< Individual >( generation.size() );
        g.addAll( generation );
        Collections.sort( g );
        return g;
    }

    public CellularGA(int sizeGeneration, double sizeElite, double probabilityMutation, IndividualFactory factory) {

        width = (int)(Math.sqrt( sizeGeneration ) + 0.5000000000001);
        height = sizeGeneration / width;

        if ( width * height != sizeGeneration )
        {
            throw new RuntimeException( "sizeGeneration: should be a square" );
        }

        this.probabilityMutation = probabilityMutation;
        this.sizeElite = sizeElite;
        generation = new ArrayList<Individual>(sizeGeneration);
        for (int i = 0; i < sizeGeneration; i++) {
            generation.add( factory.randomIndividual() );
        }
        this.factory = factory;
        r = new Random();
    }

    public void bigMutation() {
        for (int i = 0; i < generation.size(); i++) {
            generation.set(i, factory.randomIndividual());
        }
    }

    public Individual getBest()
    {
        Individual best = generation.get( 0 );
        for ( Individual i : generation )
        {
            if ( i.fitness() > best.fitness() )
            {
                best = i;
            }
        }
        return best;
    }

    private List<Individual> getNeighbours( int x, int y )
    {
        List< Individual > result = new ArrayList<Individual>();

        int[] dx = { 1, -1, 0, 0 };
        int[] dy = { 0, 0, 1, -1 };

        for ( int i = 0; i != 4; ++i )
        {
            int cur_x = ( x + dx[ i ] + height ) % height;
            int cur_y = ( y + dy[ i ] + width ) % width;

            int idx = cur_x * width + cur_y;
            result.add( generation.get( idx ) );
        }

        return result;
    }
}
