package individual.BS;

import java.util.Random;
import laboratory.common.Visualizable;
import laboratory.common.ga.Individual;
import task.ant.extended.Ant;
import task.ant.extended.ExtendedAnt;

public class BSAutomaton implements Visualizable {

    public final double mu;
    private double fitness = Double.NEGATIVE_INFINITY;
    public final BitArray autom;
    public final int statesNumber;
    public final int initialState;
    public int currentStateNumber;
    public final int ct = 256;

    public BSAutomaton( int statesNumber, int initialState, double mu ) {
        autom = new BitArray( statesNumber, ct );
        this.statesNumber = statesNumber;
        this.initialState = initialState;
        currentStateNumber = this.initialState;
        this.mu = mu;
    }

    public boolean move(ExtendedAnt ant) {
        boolean[] visible = ant.F();
        int power = 1;
        int index = 0;
        for ( boolean aVisible : visible ) {
            if ( aVisible ) {
                index += power;
            }
            power *= 2;
        }
        int nextState = autom.getNextState( currentStateNumber, index );
        Ant.Action action;
        action = autom.getAction(currentStateNumber, index);
        if ( nextState == -1 )
        {
           return false;
        }
        currentStateNumber = nextState;
        if ( action == Ant.Action.L ) {
            ant.L();
        } else if ( action == Ant.Action.R ) {
            ant.R();
        } else if ( action == Ant.Action.M ) {
            ant.M();
            if ( visible[0] ) {
                return true;
            }
        }
        return false;
    }

    public Object[] getAttributes() {
        return new Object[]{new BSAutomatonMover(this, mu)};
    }

    public double fitness() {
        if (fitness != Double.NEGATIVE_INFINITY) {
            return fitness;
        }
        fitness = 0;
        for (int j = 0; j < BSAutomatonFactory.attempts; j++) {
            ExtendedAnt ant = new ExtendedAnt(mu);
            for (int i = 0; i < Ant.NUMBER_STEPS; i++) {
                if (move(ant)) {
                    fitness++;
                }
            }
        }
        fitness /= (double) BSAutomatonFactory.attempts;
        return fitness;
    }

    public Individual mutate(Random r) {
        individual.BS.BSAutomaton ant = new individual.BS.BSAutomaton(statesNumber, r.nextInt(statesNumber), mu);

        System.arraycopy(autom.array, 0, ant.autom.array, 0, autom.array.length);
        for (int i = 2; i < statesNumber; i++) {
            for (int j = 0; j < ct; j++) {
                if (r.nextDouble() < 0.1) {
                    ant.autom.setNextState(i, j, r.nextInt(statesNumber + 1) - 1);
                }
                if (r.nextDouble() < 0.1) {
                    switch (r.nextInt(3)) {
                        case 0:
                            ant.autom.setAction(i, j, Ant.Action.L);
                            break;
                        case 1:
                            ant.autom.setAction(i, j, Ant.Action.M);
                            break;
                        case 2:
                            ant.autom.setAction(i, j, Ant.Action.R);
                            break;
                    }
                }
            }
        }

        
        return ant;
    }

    public Individual[] crossover(Individual p, Random r) {
        individual.BS.BSAutomaton[] sons = new individual.BS.BSAutomaton[2];
        individual.BS.BSAutomaton parent = (individual.BS.BSAutomaton) p;

        sons[0] = new individual.BS.BSAutomaton(statesNumber, r.nextInt(statesNumber), mu);
        sons[1] = new individual.BS.BSAutomaton(statesNumber, r.nextInt(statesNumber), mu);

        for ( ; ; ) {
            int externalNumber = r.nextInt(autom.array.length);
            int externalState = externalNumber / 2560, externalIndex = (externalNumber % 2560) / 10;

            System.arraycopy(autom.array, 0, sons[0].autom.array, 0, externalNumber);
            System.arraycopy(parent.autom.array, 0, sons[1].autom.array, 0, externalNumber);

            System.arraycopy(autom.array, externalNumber, sons[1].autom.array, externalNumber, autom.array.length - externalNumber);
            System.arraycopy(parent.autom.array, externalNumber, sons[0].autom.array, externalNumber, autom.array.length - externalNumber);

            if (sons[0].autom.getAction(externalState, externalIndex) != null && sons[1].autom.getAction(externalState, externalIndex) != null )
                break;
        }
        return sons;
    }

    public int compareTo(Individual o) {
        if (fitness() > o.fitness()) {
            return -1;
        } else {
            return 1;
        }
    }

    public String toString() {
        String res = "";
        res += "Start: "+ initialState;
        res += "\n\t";
        for (int i = 0; i < statesNumber; i++) {
            res += "  " + i+"\t";
        }
        res += "\t";

        res += "\n";

        for (int j = 0; j < ct; j++) {
            res += j+"\t";
            for (int i = 0; i < statesNumber; i++) {
                int nextState = autom.getNextState(i, j);
                if (nextState != -1)
                    res += "["+nextState+"; "+ autom.getAction(i, j)+"]\t";
                else
                    res += "   \t";
            }
            res += "\t";

            res += "\n";
        }
        return res;
    }
}
