package individual.BS;

import task.ant.extended.Ant;

public class BitArray {

    public static int NUMBER_LENGTH = 8;
    public static int ACTION_LENGTH = 2;
    private final int ct;
    private final int countStates;
    public final boolean[] array;

    public BitArray(int countStates, int ct) {
        this.ct = ct;
        this.countStates = countStates;
        array = new boolean[countStates * (NUMBER_LENGTH + ACTION_LENGTH) * ct];
    }

    public void setAction(int stateNumber, int index, Ant.Action action) {
        int begin = stateNumber * (NUMBER_LENGTH + ACTION_LENGTH) * ct + (NUMBER_LENGTH + ACTION_LENGTH) * index + NUMBER_LENGTH;
        if (action == Ant.Action.L) {
            array[begin] = false;
            array[begin + 1] = true;
        } else if (action == Ant.Action.M) {
            array[begin] = true;
            array[begin + 1] = false;
        } else if (action == Ant.Action.R) {
            array[begin] = true;
            array[begin + 1] = true;
        } else {
            array[begin] = false;
            array[begin + 1] = false;
        }
    }

    public void setNextState(int stateNumber, int index, int nextState) {
        nextState++;
        int begin = stateNumber * (NUMBER_LENGTH + ACTION_LENGTH) * ct + (NUMBER_LENGTH + ACTION_LENGTH) * index;
        for (int i = 0; i < NUMBER_LENGTH; i++) {
            if (nextState % 2 == 1) {
                array[begin + i] = true;
            } else {
                array[begin + i] = false;
            }
            nextState /= 2;
        }
    }

    public Ant.Action getAction(int stateNumber, int index) {
        int begin = stateNumber * (NUMBER_LENGTH + ACTION_LENGTH) * ct + (NUMBER_LENGTH + ACTION_LENGTH) * index + NUMBER_LENGTH;
        if (array[begin]) {
            if (array[begin + 1]) {
                return Ant.Action.R;
            } else {
                return Ant.Action.M;
            }
        } else {
            if (array[begin + 1]) {
                return Ant.Action.L;
            } else {
                return null;
            }
        }
    }

    public int getNextState(int stateNumber, int index) {
        int nextState = 0;
        int power = 1;
        int begin = stateNumber * (NUMBER_LENGTH + ACTION_LENGTH) * ct + (NUMBER_LENGTH + ACTION_LENGTH) * index;
        for (int i = 0; i < NUMBER_LENGTH; i++) {
            if (array[begin + i]) {
                nextState += power;
            }
            power *= 2;
        }
        if (nextState == 0 || nextState > countStates) {
            return -1;
        }
        nextState--;
        return nextState;
    }
}
